
/* file        : sha.c
 * created     : Thu Oct 16 15:13:19 MET DST 1997
 *
 * description : the SHA-1 message digest algorithm (Handbook Alg. 9.53) 
 * author      : Damian Weber
 */

#include <malloc.h>

#include "sha.h"

/* left rotating 32 bits */

#define ROTATE_LEFT(x,n)      ((x << n) | (x >> (32 - n)))

/* the round functions of MD4 */

inline WORD f(WORD u, WORD v, WORD w)
            {
	    return (u&v)|((~u)&w);
	    }

inline WORD g(WORD u, WORD v, WORD w)
            {
	    return (u&v)|(u&w)|(v&w);
	    }

inline WORD h(WORD u, WORD v, WORD w)
            {
	    return u^v^w;
	    }

/* byte array m[0],m[1],m[2],m[3],.... 
   is viewed as array of high endian 32-bit words

   this function converts the byte array machine-independently
   to the 32-bit-WORD array x[]
   */

void high_endian_bytes_to_words(WORD *x,BYTE *m,WORD bb)
     {
     WORD j;

     memset(x,0,bb+(4-bb%4));
     for (j=0; j<bb; j++)
	 x[j/4]|= (WORD)m[j]<<((3-(j%4))*8);
     }

/* word array w[0],w[1],w[2],w[3],.... 
   is viewed as array of high endian 32-bit words

   this function converts the word array machine-independently
   to the high-endian byte stream h
   */

void words_to_high_endian(BYTE *h,WORD *w,WORD wl)
     {
     WORD j,k;

     memset(h,0,wl*4);
     for (j=0; j<wl; j++)
         for (k=0;k<=3;k++)
	     h[4*j+k]=(BYTE)((w[j]>>((3-k)*8))&0xff);
         
     }

/* print 32-bit words x[0],...,x[l-1] in hex format */

void print_words(WORD *x, int l)
     {
     int j,k;

     for (j=0;j<l;j++)
             printf(" 0x%08x",x[j]);
     printf("\n");
     }

void sha(const BYTE *message_orig, WORD bb, BYTE *hash) /* bytelength bb is maximal 2^32-1 */
     {
     BYTE *message;
     WORD *x;
     BYTE *p;
     register int j;
     register WORD A,B,C,D,E;
     int m;
     int i;
     int rb;
     WORD bl; /* bitlength */
     WORD t;
     WORD H_ARRAY[5];

     WORD h1=0x67452301;
     WORD h2=0xefcdab89;
     WORD h3=0x98badcfe;
     WORD h4=0x10325476;
     WORD h5=0xc3d2e1f0;
     WORD y1=0x5a827999;
     WORD y2=0x6ed9eba1;
     WORD y3=0x8f1bbcdc;
     WORD y4=0xca62c1d6;
     WORD H1,H2,H3,H4,H5;

     WORD *X;

     bl=bb*8;

     /* overall processing */

     memset(hash,0,20);    /* init hash */
     message=malloc(bb+1+64+8);  /* allocate length+1+pad (max 64) + 8 byte */
     memcpy(message, message_orig,bb);

     x=malloc(bb+1+64+8);  /* allocate length+1+pad (max 64) + 8 byte */
     X=malloc(80*4);       /* allocate 80-word-block */

     rb=56-((bb+1)%64); /* pad length in bytes - multiple of 64 bytes */
     if (rb<0) rb+=64;

     message[bb]=0x80;
     for (j=1;j<=rb+4;j++) /* 0-PAD + hi-word of length */
         message[bb+j]=0;

     message[bb+rb+8]=(BYTE)(bl & 0xff);
     message[bb+rb+7]=(BYTE)((bl>>8) & 0xff);
     message[bb+rb+6]=(BYTE)((bl>>16) & 0xff);
     message[bb+rb+5]=(BYTE)((bl>>24) & 0xff);

     high_endian_bytes_to_words(x,message,bb+rb+9);

#if 0
     printf("message: pad length=%d bits\n",(bb+rb+9)*8);
     print_words(x,(bb+rb+9)/4);
#endif


     H1=h1; H2=h2; H3=h3; H4=h4; H5=h5;

     m=(bb+rb+9)/64; /* 16m 32-bit words */

     for (i=0; i<=m-1; i++)    /* processing */
         {
         for (j=0; j<=15; j++)    /* copy */
	     X[j]=x[16*i+j];
         for (j=16; j<=79; j++)    /* expand */
	     {
	     X[j]=X[j-3]^X[j-8]^X[j-14]^X[j-16];
	     X[j]=ROTATE_LEFT(X[j],1);
	     }
	 A=H1; B=H2; C=H3; D=H4; E=H5;
	 for (j=0;j<=19;j++)
	     {
	     t=ROTATE_LEFT(A,5)+f(B,C,D)+E+X[j]+y1;
	     E=D; D=C; C=ROTATE_LEFT(B,30); B=A; A=t; 
	     }
	 for (j=20;j<=39;j++)
	     {
	     t=ROTATE_LEFT(A,5)+h(B,C,D)+E+X[j]+y2;
	     E=D; D=C; C=ROTATE_LEFT(B,30); B=A; A=t; 
	     }
	 for (j=40;j<=59;j++)
	     {
	     t=ROTATE_LEFT(A,5)+g(B,C,D)+E+X[j]+y3;
	     E=D; D=C; C=ROTATE_LEFT(B,30); B=A; A=t; 
	     }
	 for (j=60;j<=79;j++)
	     {
	     t=ROTATE_LEFT(A,5)+h(B,C,D)+E+X[j]+y4;
	     E=D; D=C; C=ROTATE_LEFT(B,30); B=A; A=t; 
	     }
	 H1+=A; H2+=B; H3+=C; H4+=D; H5+=E;
	 }

     H_ARRAY[0]=H1;
     H_ARRAY[1]=H2;
     H_ARRAY[2]=H3;
     H_ARRAY[3]=H4;
     H_ARRAY[4]=H5;

     words_to_high_endian(hash,H_ARRAY,5);
     free(x); free(X);
     }
      
     

/* compute hash of secret1|message_orig|secret2, where secret1/2 are the two 
 * halves of secret */

void sha_hmac_kxk(const BYTE *secret, const int secret_length, const BYTE *message_orig, WORD bb, BYTE *hash) /* bytelength bb is maximal 2^32-1 */
{
	char *message_tmp;
	int  j,offset,sl2;

	sl2=secret_length/2;
	message_tmp=malloc(bb+secret_length);

	for (j=0;j<sl2;j++)
		message_tmp[j]=secret[j];

	for (j=0;j<bb;j++)
		message_tmp[j+sl2]=message_orig[j];

	offset=sl2+bb;
	for (j=0;j<sl2;j++)
		message_tmp[j+offset]=secret[j+sl2];

	if (secret_length % 2 != 0)
		message_tmp[j+offset]=secret[secret_length-1];

	sha(message_tmp,bb+secret_length,hash);
}

